@@ -6,7 +6,7 @@
 #ifndef _IP_VS_H
 #define _IP_VS_H
 
-#define IP_VS_VERSION_CODE            0x000205
+#define IP_VS_VERSION_CODE            0x000206
 #define NVERSION(version)                       \
         (version >> 16) & 0xFF,                 \
         (version >> 8) & 0xFF,                  \
@@ -83,6 +83,9 @@ struct ip_vs_rule_user {
 #define IP_VS_SO_GET_MAX	IP_VS_SO_GET_VERSION
 
 
+/*
+ *      IPVS Connection Flags
+ */
 #define IP_VS_CONN_F_FWD_MASK         0x0003    /* mask for the fwd methods */
 #define IP_VS_CONN_F_MASQ	      0x0000    /* masquerading */
 #define IP_VS_CONN_F_LOCALNODE	      0x0001    /* local node */
@@ -145,15 +148,16 @@ extern int ip_vs_get_debug_level(void);
 #define LeaveFunction(level)   do {} while (0)
 #endif
 
+
 /*
- *  The port number of FTP service (in network order).
+ *      The port number of FTP service (in network order).
  */
 #define FTPPORT  __constant_htons(21)
 #define FTPDATA  __constant_htons(20)
 
 
 /*
- *  IPVS sysctl variables under the /proc/sys/net/ipv4/vs/
+ *      IPVS sysctl variables under the /proc/sys/net/ipv4/vs/
  */
 #define NET_IPV4_VS              21
 
@@ -182,6 +186,10 @@ enum {
 	NET_IPV4_VS_CACHE_BYPASS=22,
 };
 
+
+/*
+ *      IPVS State Values
+ */
 enum {
 	IP_VS_S_NONE = 0,
 	IP_VS_S_ESTABLISHED,
@@ -228,6 +236,20 @@ struct ip_vs_seq {
 };
 
 
+/*
+ *	IPVS statistics object
+ */
+struct ip_vs_stats
+{
+        spinlock_t              lock;           /* spin lock */
+        unsigned long           conns;          /* connections scheduled */
+        unsigned long           inpkts;         /* incoming packets */
+        unsigned long           outpkts;        /* outgoing packets */
+        unsigned long           inbytes;        /* incoming bytes */
+        unsigned long           outbytes;       /* outgoing bytes */
+};
+
+
 /*
  *	IP_VS structure allocated for each dynamically scheduled connection
  */
@@ -269,8 +291,8 @@ struct ip_vs_conn {
  *	and the forwarding entries
  */
 struct ip_vs_service {
-	struct list_head	s_list;   /* hashed d-linked list head */
-	struct list_head	f_list;   /* hashed d-linked list head */
+	struct list_head	s_list;   /* for normal service table */
+	struct list_head	f_list;   /* for fwmark-based service table */
 	atomic_t		refcnt;   /* reference counter */
 	__u16			protocol; /* which protocol (TCP/UDP) */
 	__u32			addr;	  /* IP address for virtual service */
@@ -283,7 +305,7 @@ struct ip_vs_service {
 	struct ip_vs_scheduler	*scheduler;    /* bound scheduler object */
 	rwlock_t		sched_lock;    /* lock sched_data */
 	void			*sched_data;   /* scheduler application data */
-	atomic_t		packethandled; /* packet handled */
+        struct ip_vs_stats      stats;         /* statistics for the service */
 };
 
 
@@ -292,22 +314,27 @@ struct ip_vs_service {
  *	with ip address, port number, and so on.
  */
 struct ip_vs_dest {
-	struct list_head	n_list;   /* d-linked list head */
+	struct list_head	n_list;   /* for the dests in the service */
+	struct list_head	d_list;   /* for table with all the dests */
+
 	__u32			addr;	  /* IP address of real server */
 	__u16			port;	  /* port number of the service */
 	unsigned		flags;	  /* dest status flags */
 	atomic_t		weight; 	/* server weight */
-	atomic_t		conn_flags;	/* flags to copy to masq */
+	atomic_t		conn_flags;	/* flags to copy to conn */
 	atomic_t		activeconns;	/* active connections */
 	atomic_t		inactconns;	/* inactive connections */
 	atomic_t		refcnt; 	/* reference counter */
-	struct list_head	d_list;         /* table with all dests */
+        struct ip_vs_stats      stats;          /* statistics */
+
+        /* for destination cache */
 	spinlock_t		dst_lock;	/* lock dst_cache */
 	struct dst_entry	*dst_cache;	/* destination cache entry */
 	u32			dst_rtos;	/* RT_TOS(tos) for dst */
 
 	/* for virtual service */
-	__u16			protocol; /* which protocol (TCP/UDP) */
+	struct ip_vs_service    *svc;     /* service that it belongs to */
+        __u16			protocol; /* which protocol (TCP/UDP) */
 	__u32			vaddr;	  /* IP address for virtual service */
 	__u16			vport;	  /* port number for the service */
 	unsigned long		vfwmark;  /* firewall mark of the service */
@@ -359,7 +386,7 @@ struct ip_vs_app
                       struct ip_vs_conn *, struct sk_buff **);
 };
 
-
+        
 /*
  *      IPVS core functions
  *      (from ip_vs_core.c)
@@ -604,14 +631,17 @@ __ip_vs_dst_reset(struct ip_vs_dest *dest)
 static inline struct dst_entry *
 __ip_vs_dst_check(struct ip_vs_dest *dest, u32 rtos, u32 cookie)
 {
-	struct dst_entry *dst = dest->dst_cache;
+        struct dst_entry *dst = dest->dst_cache;
 
-	if (dst && (dst->obsolete || rtos != dest->dst_rtos) &&
-		dst->ops->check(dst, cookie) == NULL) {
-		dest->dst_cache = 0;
-		return NULL;
-	}
-	return dst;
+        if (!dst)
+                return NULL;
+        if ((dst->obsolete || rtos != dest->dst_rtos) &&
+            dst->ops->check(dst, cookie) == NULL) {
+                dest->dst_cache = 0;
+                return NULL;
+        }
+        dst_hold(dst);
+        return dst;
 }
 
 
