@@ -1,7 +1,7 @@
 /*
  * IPVS         Application module
  *
- * Version:     $Id: ip_vs_app.c,v 1.7 2001/02/23 09:15:05 wensong Exp $
+ * Version:     $Id: ip_vs_app.c,v 1.8 2001/03/18 15:01:45 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -414,46 +414,48 @@ int ip_vs_app_pkt_in(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 static int ip_vs_app_getinfo(char *buffer, char **start, off_t offset,
 			     int length)
 {
-        off_t pos=0, begin;
+        off_t pos=0;
         int len=0;
-        struct ip_vs_app * vapp;
-        int idx, size;
+        char temp[64];
+        int idx;
+        struct ip_vs_app *vapp;
         struct list_head *e;
 
-        size = sprintf(buffer, "prot port    usecnt name\n");
-        pos += size;
-        len += size;
+        pos = 64;
+        if (pos > offset) {
+                len += sprintf(buffer+len, "%-63s\n",
+                               "prot port    usecnt name");
+        }
 
+        read_lock_bh(&__ip_vs_app_lock);
         for (idx=0 ; idx < IP_VS_APP_TAB_SIZE; idx++) {
-                read_lock_bh(&__ip_vs_app_lock);
                 list_for_each (e, &ip_vs_app_base[idx]) {
                         vapp = list_entry(e, struct ip_vs_app, n_list);
 
-                        size += sprintf(buffer+len, "%-3s  %-7u %-6d %-17s\n",
-                                        ip_vs_proto_name(
-                                                IP_VS_APP_PROTO(vapp->type)),
-                                        IP_VS_APP_PORT(vapp->type),
-                                        GET_USE_COUNT(vapp->module),
-                                        vapp->name);
-
-                        len += size;
-                        pos += size;
+                        pos += 64;
                         if (pos <= offset)
-                                len=0;
-                        if (pos >= offset+length) {
-                                read_unlock_bh(&__ip_vs_app_lock);
+                                continue;
+                        if (pos > offset+length) {
+                                pos -= 64;
                                 goto done;
                         }
+                        sprintf(temp, "%-3s  %-7u %-6d %-17s",
+                                ip_vs_proto_name(IP_VS_APP_PROTO(vapp->type)),
+                                IP_VS_APP_PORT(vapp->type),
+                                GET_USE_COUNT(vapp->module),
+                                vapp->name);
+                        len += sprintf(buffer+len, "%-63s\n", temp);
                 }
-                read_unlock_bh(&__ip_vs_app_lock);
         }
-
   done:
-	begin = len - (pos - offset);
-	*start = buffer + begin;
-	len -= begin;
-	if(len>length)
-		len = length;
+        read_unlock_bh(&__ip_vs_app_lock);
+
+        *start = buffer+len-(pos-offset);       /* Start of wanted data */
+        len = pos-offset;
+        if (len > length)
+                len = length;
+        if (len < 0)
+                len = 0;
 	return len;
 }
 
