@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_conn.c,v 1.11 2001/01/26 10:19:22 wensong Exp $
+ * Version:     $Id: ip_vs_conn.c,v 1.12 2001/03/18 15:01:45 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -1317,16 +1317,17 @@ ip_vs_conn_new(int proto, __u32 caddr, __u16 cport, __u32 vaddr, __u16 vport,
 static int
 ip_vs_conn_getinfo(char *buffer, char **start, off_t offset, int length)
 {
-	off_t pos=0, begin;
+	off_t pos=0;
+        int idx, len=0;
+        char temp[70];
 	struct ip_vs_conn *cp;
-        int idx, size;
-	int len=0;
 	struct list_head *l, *e;
 
-        size = sprintf(buffer, "Prc FromIP   FPrt ToIP     TPrt "
-                       "DestIP   DPrt State       Expires\n");
-        pos += size;
-        len += size;
+        pos = 70;
+        if (pos > offset) {
+                len += sprintf(buffer+len, "%-69s\n",
+                               "Pro FromIP   FPrt ToIP     TPrt DestIP   DPrt State       Expires");
+        }
 
         for(idx = 0; idx < IP_VS_CONN_TAB_SIZE; idx++) {
                 /*
@@ -1338,34 +1339,35 @@ ip_vs_conn_getinfo(char *buffer, char **start, off_t offset, int length)
                 l = &ip_vs_table[idx];
                 for (e=l->next; e!=l; e=e->next) {
                         cp = list_entry(e, struct ip_vs_conn, c_list);
-
-                        size = sprintf(buffer+len,
-                                       "%-3s %08X %04X %08X %04X "
-                                       "%08X %04X %-11s %7lu\n",
-                                       ip_vs_proto_name(cp->protocol),
-                                       ntohl(cp->caddr), ntohs(cp->cport),
-                                       ntohl(cp->vaddr), ntohs(cp->vport),
-                                       ntohl(cp->daddr), ntohs(cp->dport),
-                                       ip_vs_state_name(cp->state),
-                                       cp->timer.expires-jiffies);
-                        len += size;
-                        pos += size;
+                        pos += 70;
                         if (pos <= offset)
-                                len=0;
-                        if (pos >= offset+length) {
+                                continue;
+                        if (pos > offset+length) {
+                                pos -= 70;
                                 ct_read_unlock_bh(idx);
                                 goto done;
                         }
+
+                        sprintf(temp,
+                                "%-3s %08X %04X %08X %04X %08X %04X %-11s %7lu",
+                                ip_vs_proto_name(cp->protocol),
+                                ntohl(cp->caddr), ntohs(cp->cport),
+                                ntohl(cp->vaddr), ntohs(cp->vport),
+                                ntohl(cp->daddr), ntohs(cp->dport),
+                                ip_vs_state_name(cp->state),
+                                cp->timer.expires-jiffies);
+                        len += sprintf(buffer+len, "%-69s\n", temp);
                 }
                 ct_read_unlock_bh(idx);
 	}
 
   done:
-	begin = len - (pos - offset);
-	*start = buffer + begin;
-	len -= begin;
-	if(len>length)
-		len = length;
+        *start = buffer+len-(pos-offset);       /* Start of wanted data */
+        len = pos-offset;
+        if (len > length)
+                len = length;
+        if (len < 0)
+                len = 0;
 	return len;
 }
 
