@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_core.c,v 1.14 2001/02/23 09:15:05 wensong Exp $
+ * Version:     $Id: ip_vs_core.c,v 1.15 2001/03/18 15:01:45 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -132,6 +132,75 @@ const char *ip_vs_proto_name(unsigned proto)
         }
 }
 
+extern struct ip_vs_stats ip_vs_stats;
+extern rwlock_t __ip_vs_svc_lock;
+
+static inline void
+ip_vs_in_stats(struct ip_vs_conn *cp, struct sk_buff *skb)
+{
+        struct ip_vs_dest *dest = cp->dest;
+        read_lock(&__ip_vs_svc_lock);
+        if (dest && (dest->flags & IP_VS_DEST_F_AVAILABLE)) {
+                spin_lock(&dest->stats.lock);
+                dest->stats.inpkts++;
+                dest->stats.inbytes += skb->len;
+                spin_unlock(&dest->stats.lock);
+
+                spin_lock(&dest->svc->stats.lock);
+                dest->svc->stats.inpkts++;
+                dest->svc->stats.inbytes += skb->len;
+                spin_unlock(&dest->svc->stats.lock);
+
+                spin_lock(&ip_vs_stats.lock);
+                ip_vs_stats.inpkts++;
+                ip_vs_stats.inbytes += skb->len;
+                spin_unlock(&ip_vs_stats.lock);
+        }
+        read_unlock(&__ip_vs_svc_lock);
+}
+
+
+static inline void
+ip_vs_out_stats(struct ip_vs_conn *cp, struct sk_buff *skb)
+{
+        struct ip_vs_dest *dest = cp->dest;
+        read_lock(&__ip_vs_svc_lock);
+        if (dest && (dest->flags & IP_VS_DEST_F_AVAILABLE)) {
+                spin_lock(&dest->stats.lock);
+                dest->stats.outpkts++;
+                dest->stats.outbytes += skb->len;
+                spin_unlock(&dest->stats.lock);
+
+                spin_lock(&dest->svc->stats.lock);
+                dest->svc->stats.outpkts++;
+                dest->svc->stats.outbytes += skb->len;
+                spin_unlock(&dest->svc->stats.lock);
+
+                spin_lock(&ip_vs_stats.lock);
+                ip_vs_stats.outpkts++;
+                ip_vs_stats.outbytes += skb->len;
+                spin_unlock(&ip_vs_stats.lock);
+        }
+        read_unlock(&__ip_vs_svc_lock);
+}
+
+
+static inline void
+ip_vs_conn_stats(struct ip_vs_conn *cp, struct ip_vs_service *svc)
+{
+        spin_lock(&cp->dest->stats.lock);
+        cp->dest->stats.conns++;
+        spin_unlock(&cp->dest->stats.lock);
+
+        spin_lock(&svc->stats.lock);
+        svc->stats.conns++;
+        spin_unlock(&svc->stats.lock);
+
+        spin_lock(&ip_vs_stats.lock);
+        ip_vs_stats.conns++;
+        spin_unlock(&ip_vs_stats.lock);
+}
+
 
 /*
  *	Own skb_cow() beast, tweaked for rewriting commonly
@@ -469,6 +538,9 @@ static int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
 
                 /* transmit the first SYN packet */
                 ret = cp->packet_xmit(skb, cp);
+
+                /* statistics */
+                ip_vs_in_stats(cp, skb);
                 
                 ip_vs_conn_put(cp);
                 return ret;
@@ -608,6 +680,8 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 
 	/* the TCP/UDP dest port - cannot redo check */
 	pptr[1] = cp->vport;
+
+        ip_vs_out_stats(cp, skb);
 	__ip_vs_conn_put(cp);
 
 	/* And finally the ICMP checksum */
@@ -805,6 +879,7 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	}
 	ip_send_check(iph);
 
+        ip_vs_out_stats(cp, skb);
 	ip_vs_set_state(cp, VS_STATE_OUTPUT, iph, h.portp);
 	ip_vs_conn_put(cp);
         
@@ -899,13 +974,15 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
                   NIPQUAD(ciph->daddr), ntohs(pptr[1]));
 
 	/* This is pretty much what ip_vs_conn_in_get() does,
-           except parameters are wrong way round */
+           except parameters are in the reverse order */
         cp = ip_vs_conn_in_get(ciph->protocol,
                                ciph->daddr, pptr[1],
                                ciph->saddr, pptr[0]);
 	if (cp == NULL)
 		return NF_ACCEPT;
 
+        ip_vs_in_stats(cp, skb);
+
         /* The ICMP packet for VS/TUN, VS/DR and LOCALNODE will be
            forwarded directly here, because there is no need to
            translate address/port back */
@@ -1096,12 +1173,14 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 		cp = ip_vs_schedule(svc, iph);
 		if (!cp)
 			return ip_vs_leave(svc, skb);
+                ip_vs_conn_stats(cp, svc);
                 ip_vs_service_put(svc);
 	}
 
 	if (cp != NULL) {
                 int ret;
 
+                ip_vs_in_stats(cp, skb);
                 ip_vs_set_state(cp, VS_STATE_INPUT, iph, h.portp); 
                 if (cp->packet_xmit)
                         ret = cp->packet_xmit(skb, cp);
