@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_ctl.c,v 1.13 2001/02/15 15:17:59 wensong Exp $
+ * Version:     $Id: ip_vs_ctl.c,v 1.14 2001/03/18 15:01:45 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -255,19 +255,21 @@ static int ip_vs_svc_hash(struct ip_vs_service *svc)
 
 	if (svc->fwmark == 0) {
 		/*
-		 *	Hash by <protocol,addr,port> in ip_vs_svc_table
+		 *  Hash it by <protocol,addr,port> in ip_vs_svc_table
 		 */
 		hash = ip_vs_svc_hashkey(svc->protocol, svc->addr, svc->port);
 		list_add(&svc->s_list, &ip_vs_svc_table[hash]);
 	} else {
 		/*
-		 *	Hash by fwmark in ip_vs_svc_fwm_table
+		 *  Hash it by fwmark in ip_vs_svc_fwm_table
 		 */
 		hash = ip_vs_svc_fwm_hashkey(svc->fwmark);
 		list_add(&svc->f_list, &ip_vs_svc_fwm_table[hash]);
 	}
 	
 	svc->flags |= IP_VS_SVC_F_HASHED;
+        /* increase its refcnt because it is referenced by the svc table */
+        atomic_inc(&svc->refcnt);
 	return 1;
 }
 
@@ -298,6 +300,7 @@ static int ip_vs_svc_unhash(struct ip_vs_service *svc)
 	}
 
 	svc->flags &= ~IP_VS_SVC_F_HASHED;
+        atomic_dec(&svc->refcnt);
 	return 1;
 }
 
@@ -637,6 +640,7 @@ static int __ip_vs_update_dest(struct ip_vs_service *svc,
 	 *    Set the dest status flags
 	 */
 	dest->flags |= IP_VS_DEST_F_AVAILABLE;
+	dest->svc = svc;
 
 	return 0;
 }
@@ -662,7 +666,7 @@ ip_vs_new_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur,
 	}
 	memset(dest, 0, sizeof(struct ip_vs_dest));
 
-	dest->protocol = svc->protocol;
+        dest->protocol = svc->protocol;
 	dest->vaddr = svc->addr;
 	dest->vport = svc->port;
 	dest->vfwmark = svc->fwmark;
@@ -675,6 +679,7 @@ ip_vs_new_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur,
 
 	INIT_LIST_HEAD(&dest->d_list);
 	dest->dst_lock = SPIN_LOCK_UNLOCKED;
+        dest->stats.lock = SPIN_LOCK_UNLOCKED;
 	ret = __ip_vs_update_dest(svc, dest, ur);
 	if (ret) kfree(dest);
 
@@ -737,9 +742,10 @@ static int ip_vs_add_dest(struct ip_vs_service *svc,
 		write_lock_bh(&__ip_vs_svc_lock);
 
 		/*
-		 * Wait until all svc users go away
+		 * Wait until all other svc users go away, and only the
+                 * table and I are the referrers. 
 		 */
-		while (atomic_read(&svc->refcnt) > 1) {};
+		while (atomic_read(&svc->refcnt) > 2) {};
 
 		list_add(&dest->n_list, &svc->destinations);
 
@@ -766,9 +772,10 @@ static int ip_vs_add_dest(struct ip_vs_service *svc,
 	write_lock_bh(&__ip_vs_svc_lock);
 
 	/*
-	 *	Wait until all svc users go away
+	 * Wait until all other svc users go away, only the table and I are
+         * the referrers. 
 	 */
-	while (atomic_read(&svc->refcnt) > 1) {};
+	while (atomic_read(&svc->refcnt) > 2) {};
 
 	list_add(&dest->n_list, &svc->destinations);
 
@@ -871,6 +878,9 @@ static void __ip_vs_unlink_dest(struct ip_vs_service *svc,
 }
 
 
+/*
+ *  Delete a destination server in the given service
+ */
 static int ip_vs_del_dest(struct ip_vs_service *svc,struct ip_vs_rule_user *ur)
 {
 	struct ip_vs_dest *dest;
@@ -879,9 +889,6 @@ static int ip_vs_del_dest(struct ip_vs_service *svc,struct ip_vs_rule_user *ur)
 
 	EnterFunction(2);
 
-	/*
-	 *    Lookup the destination list
-	 */
 	dest = ip_vs_lookup_dest(svc, daddr, dport);
 	if (dest == NULL) {
 		IP_VS_DBG(1, "ip_vs_del_dest(): destination not found!\n");
@@ -891,14 +898,16 @@ static int ip_vs_del_dest(struct ip_vs_service *svc,struct ip_vs_rule_user *ur)
 	write_lock_bh(&__ip_vs_svc_lock);
 
 	/*
-	 *	Wait until all svc users go away
+	 *	Wait until all other svc users go away, only the table
+         *      and I are the referrers.
 	 */
-	while (atomic_read(&svc->refcnt) > 1) {};
+	while (atomic_read(&svc->refcnt) > 2) {};
 
 	/*
 	 *	Unlink dest from the service
 	 */
 	__ip_vs_unlink_dest(svc, dest, 1);
+        
 	write_unlock_bh(&__ip_vs_svc_lock);
 
 	/*
@@ -915,7 +924,8 @@ static int ip_vs_del_dest(struct ip_vs_service *svc,struct ip_vs_rule_user *ur)
 /*
  *  Add a service into the service hash table
  */
-static int ip_vs_add_service(struct ip_vs_rule_user *ur)
+static int
+ip_vs_add_service(struct ip_vs_rule_user *ur, struct ip_vs_service **svc_p)
 {
 	__u16  protocol = ur->protocol;
 	__u32 vaddr = ur->vaddr;
@@ -924,7 +934,7 @@ static int ip_vs_add_service(struct ip_vs_rule_user *ur)
 
 	int ret = 0;
 	struct ip_vs_scheduler *sched;
-	struct ip_vs_service *svc = NULL;
+	struct ip_vs_service *svc;
 
 	MOD_INC_USE_COUNT;
 
@@ -944,7 +954,7 @@ static int ip_vs_add_service(struct ip_vs_rule_user *ur)
 	if (svc == NULL) {
 		MOD_DEC_USE_COUNT;
 		IP_VS_DBG(1, "ip_vs_add_service: kmalloc failed.\n");
-		ret =  -ENOMEM;
+		ret = -ENOMEM;
 		goto out_err;
 	}
 	memset(svc, 0, sizeof(struct ip_vs_service));
@@ -959,7 +969,8 @@ static int ip_vs_add_service(struct ip_vs_rule_user *ur)
 
 	INIT_LIST_HEAD(&svc->destinations);
 	svc->sched_lock = RW_LOCK_UNLOCKED;
-
+        svc->stats.lock = SPIN_LOCK_UNLOCKED;
+        
 	/*
 	 *    Bind the scheduler
 	 */
@@ -969,9 +980,9 @@ static int ip_vs_add_service(struct ip_vs_rule_user *ur)
 	}
 
 	/*
-	 *    First svc user: the table with the virtual services
+	 *    I'm the first user of the service
 	 */
-	atomic_set(&svc->refcnt,1);
+	atomic_set(&svc->refcnt, 1);
 
 	/*
 	 *    Hash the service into the service table
@@ -988,6 +999,7 @@ static int ip_vs_add_service(struct ip_vs_rule_user *ur)
         else if (vport == 0)
                 atomic_inc(&ip_vs_nullsvc_counter);
 
+        *svc_p = svc;
 	return 0;
 
   out_err:
@@ -1022,9 +1034,10 @@ static int ip_vs_edit_service(struct ip_vs_service *svc,
         write_lock_bh(&__ip_vs_svc_lock);
 
         /*
-         * Wait until all svc users go away
+         * Wait until all other svc users go away, and only the table
+         * and I are the referrers.
          */
-        while (atomic_read(&svc->refcnt) > 1) {};
+        while (atomic_read(&svc->refcnt) > 2) {};
 
         /*
          * Set the flags and timeout value
@@ -1083,7 +1096,7 @@ static int __ip_vs_del_service(struct ip_vs_service *svc)
         struct ip_vs_scheduler *old_sched;
 
         /*
-         * Unbind scheduler
+         *    Unbind scheduler
          */
         old_sched = svc->scheduler;
         ip_vs_unbind_scheduler(svc);
@@ -1121,6 +1134,8 @@ static int __ip_vs_del_service(struct ip_vs_service *svc)
  */
 static int ip_vs_del_service(struct ip_vs_service *svc)
 {
+        int ret;
+        
         if (svc == NULL)
                 return -EEXIST;
 
@@ -1128,16 +1143,19 @@ static int ip_vs_del_service(struct ip_vs_service *svc)
          * Unhash it from the service table
          */
         write_lock_bh(&__ip_vs_svc_lock);
+        
         ip_vs_svc_unhash(svc);
 
         /*
-         * Wait until all svc users go away
+         * Wait until all other svc users go away and only I are the referrer.
          */
         while (atomic_read(&svc->refcnt) > 1) {};
 
+        ret = __ip_vs_del_service(svc);
+
         write_unlock_bh(&__ip_vs_svc_lock);
 
-        return __ip_vs_del_service(svc);
+        return ret;
 }
 
 
@@ -1160,7 +1178,8 @@ static int ip_vs_flush(void)
                         write_lock_bh(&__ip_vs_svc_lock);
                         ip_vs_svc_unhash(svc);
                         /*
-                         *	Wait until all svc users go away
+                         * Wait until all other svc users go away and only
+                         * I are the referrer.
                          */
                         while (atomic_read(&svc->refcnt) > 1) {};
                         __ip_vs_del_service(svc);
@@ -1178,7 +1197,8 @@ static int ip_vs_flush(void)
                         write_lock_bh(&__ip_vs_svc_lock);
                         ip_vs_svc_unhash(svc);
                         /*
-                         * Wait until all svc users go away
+                         * Wait until all other svc users go away and only
+                         * I am the referrer.
                          */
                         while (atomic_read(&svc->refcnt) > 1) {};
                         __ip_vs_del_service(svc);
@@ -1323,22 +1343,31 @@ static inline char *ip_vs_fwd_name(unsigned flags)
 
 static int ip_vs_get_info(char *buf, char **start, off_t offset, int length)
 {
-        int idx;
         int len=0;
         off_t pos=0;
-        int size;
+        char temp[64], temp2[32];
+        int idx;
         struct ip_vs_service *svc;
         struct ip_vs_dest *dest;
         struct list_head *l, *e, *p, *q;
 
-        size = sprintf(buf+len,
-                       "IP Virtual Server version %d.%d.%d (size=%d)\n"
-                       "Prot LocalAddress:Port Scheduler Flags\n"
-                       "  -> RemoteAddress:Port Forward Weight ActiveConn InActConn\n",
-                       NVERSION(IP_VS_VERSION_CODE), IP_VS_CONN_TAB_SIZE);
-        pos += size;
-        len += size;
-        
+        /*
+         * Note: since the length of the buffer is usually the multiple
+         * of 512, it is easier and faster to use fixed record of the
+         * divisor of 512, in order to avoid buffer overflow. 
+         */
+        pos = 192;
+        if (pos > offset) {
+                sprintf(temp,
+                        "IP Virtual Server version %d.%d.%d (size=%d)",
+                        NVERSION(IP_VS_VERSION_CODE), IP_VS_CONN_TAB_SIZE);
+                len += sprintf(buf+len, "%-63s\n", temp);
+                len += sprintf(buf+len, "%-63s\n",
+                               "Prot LocalAddress:Port Scheduler Flags");
+                len += sprintf(buf+len, "%-63s\n",
+                               "  -> RemoteAddress:Port Forward Weight ActiveConn InActConn");
+        }
+
         read_lock_bh(&__ip_vs_svc_lock);
 
         /* print the service table hashed by <protocol,addr,port> */
@@ -1346,47 +1375,41 @@ static int ip_vs_get_info(char *buf, char **start, off_t offset, int length)
                 l = &ip_vs_svc_table[idx];
                 for (e=l->next; e!=l; e=e->next) {
                         svc = list_entry(e, struct ip_vs_service, s_list);
-                        size = sprintf(buf+len, "%s  %08X:%04X %s",
-                                       ip_vs_proto_name(svc->protocol),
-                                       ntohl(svc->addr),
-                                       ntohs(svc->port),
-                                       svc->scheduler->name);
-                        len += size;
-                        pos += size;
-                        
-                        if (svc->flags & IP_VS_SVC_F_PERSISTENT)
-                                size = sprintf(buf+len,
-                                               " persistent %d %08X\n",
-                                               svc->timeout,
-                                               ntohl(svc->netmask));
-                        else
-                                size = sprintf(buf+len, "\n");
-                        
-                        len += size;
-                        pos += size;
+                        pos += 64;
+                        if (pos > offset) {
+                                if (svc->flags & IP_VS_SVC_F_PERSISTENT)
+                                        sprintf(temp2, " persistent %d %08X\n",
+                                                svc->timeout,
+                                                ntohl(svc->netmask));
+                                else
+                                        sprintf(temp2, "\n");
+                                
+                                sprintf(temp, "%s  %08X:%04X %s",
+                                        ip_vs_proto_name(svc->protocol),
+                                        ntohl(svc->addr),
+                                        ntohs(svc->port),
+                                        svc->scheduler->name, temp2);
+                                len += sprintf(buf+len, "%-63s\n", temp);
+                                if (len >= length)
+                                        goto done;
+                        }
 
-                        if (pos <= offset)
-                                len=0;
-                        if (pos >= offset+length)
-                                goto done;
-			       
                         p = &svc->destinations;
                         for (q=p->next; q!=p; q=q->next) {
-                                dest = list_entry(q,struct ip_vs_dest,n_list);
-                                size = sprintf(buf+len,
-                                               "  -> %08X:%04X      %-7s %-6d %-10d %-10d\n",
-                                               ntohl(dest->addr),
-                                               ntohs(dest->port),
-                                               ip_vs_fwd_name(atomic_read(&dest->conn_flags)),
-                                               atomic_read(&dest->weight),
-                                               atomic_read(&dest->activeconns),
-                                               atomic_read(&dest->inactconns));
-                                len += size;
-                                pos += size;
-                                
+                                dest = list_entry(q, struct ip_vs_dest, n_list);
+                                pos += 64;
                                 if (pos <= offset)
-                                        len=0;
-                                if (pos >= offset+length)
+                                        continue;
+                                sprintf(temp,
+                                        "  -> %08X:%04X      %-7s %-6d %-10d %-10d",
+                                        ntohl(dest->addr),
+                                        ntohs(dest->port),
+                                        ip_vs_fwd_name(atomic_read(&dest->conn_flags)),
+                                        atomic_read(&dest->weight),
+                                        atomic_read(&dest->activeconns),
+                                        atomic_read(&dest->inactconns));
+                                len += sprintf(buf+len, "%-63s\n", temp);
+                                if (len >= length)
                                         goto done;
                         }
                 }
@@ -1397,44 +1420,39 @@ static int ip_vs_get_info(char *buf, char **start, off_t offset, int length)
                 l = &ip_vs_svc_fwm_table[idx];
                 for (e=l->next; e!=l; e=e->next) {
                         svc = list_entry(e, struct ip_vs_service, f_list);
-                        size = sprintf(buf+len, "%s  %08lX %s", "FWM",
-                                       svc->fwmark, svc->scheduler->name);
-                        len += size;
-                        pos += size;
-                        
-                        if (svc->flags & IP_VS_SVC_F_PERSISTENT)
-                                size = sprintf(buf+len,
-                                               " persistent %d %08X\n",
-                                               svc->timeout,
-                                               ntohl(svc->netmask));
-                        else
-                                size = sprintf(buf+len, "\n");
-                        
-                        len += size;
-                        pos += size;
+                        pos += 64;
+                        if (pos > offset) {
+                                if (svc->flags & IP_VS_SVC_F_PERSISTENT)
+                                        sprintf(temp2, " persistent %d %08X\n",
+                                                svc->timeout,
+                                                ntohl(svc->netmask));
+                                else
+                                        sprintf(temp2, "\n");
+                                
+                                sprintf(temp, "FWM  %08X %s %s",
+                                        svc->fwmark,
+                                        svc->scheduler->name, temp2);
+                                len += sprintf(buf+len, "%-63s\n", temp);
+                                if (len >= length)
+                                        goto done;
+                        }
 
-                        if (pos <= offset)
-                                len=0;
-                        if (pos >= offset+length)
-                                goto done;
-			       
                         p = &svc->destinations;
                         for (q=p->next; q!=p; q=q->next) {
-                                dest = list_entry(q,struct ip_vs_dest,n_list);
-                                size = sprintf(buf+len,
-                                               "  -> %08X:%04X      %-7s %-6d %-10d %-10d\n",
-                                               ntohl(dest->addr),
-                                               ntohs(dest->port),
-                                               ip_vs_fwd_name(atomic_read(&dest->conn_flags)),
-                                               atomic_read(&dest->weight),
-                                               atomic_read(&dest->activeconns),
-                                               atomic_read(&dest->inactconns));
-                                len += size;
-                                pos += size;
-                                
+                                dest = list_entry(q, struct ip_vs_dest, n_list);
+                                pos += 64;
                                 if (pos <= offset)
-                                        len=0;
-                                if (pos >= offset+length)
+                                        continue;
+                                sprintf(temp,
+                                        "  -> %08X:%04X      %-7s %-6d %-10d %-10d",
+                                        ntohl(dest->addr),
+                                        ntohs(dest->port),
+                                        ip_vs_fwd_name(atomic_read(&dest->conn_flags)),
+                                        atomic_read(&dest->weight),
+                                        atomic_read(&dest->activeconns),
+                                        atomic_read(&dest->inactconns));
+                                len += sprintf(buf+len, "%-63s\n", temp);
+                                if (len >= length)
                                         goto done;
                         }
                 }
@@ -1449,7 +1467,183 @@ static int ip_vs_get_info(char *buf, char **start, off_t offset, int length)
                 len = length;
         if (len < 0)
                 len = 0;
+        return len;
+}
+
+
+struct ip_vs_stats ip_vs_stats = {SPIN_LOCK_UNLOCKED, 0, 0};
+
+static int
+ip_vs_stats_get_info(char *buf, char **start, off_t offset, int length)
+{
+        int idx;
+        int len=0;
+        off_t pos=0;
+        char temp[128];
+        struct ip_vs_service *svc;
+        struct ip_vs_dest *dest;
+        struct list_head *l, *e, *p, *q;
+
+        pos += 128;
+        if (pos > offset) {
+                len += sprintf(buf+len, "%-63s\n",
+/*                              0123456789 0123456789 0123456789 0123456789 0123456789 */
+                               "TotalConns     InPkts    OutPkts    InBytes   OutBytes");
+                spin_lock(&ip_vs_stats.lock);
+                sprintf(temp, "%10lu %10lu %10lu %10lu %10lu",
+                        ip_vs_stats.conns,
+                        ip_vs_stats.inpkts,
+                        ip_vs_stats.outpkts,
+                        ip_vs_stats.inbytes,
+                        ip_vs_stats.outbytes);
+                spin_unlock(&ip_vs_stats.lock);
+                len += sprintf(buf+len, "%-63s\n", temp);
+        }
         
+        read_lock_bh(&__ip_vs_svc_lock);
+        
+        /* print the service statistics */
+        pos += 90;
+        if (pos > offset) {
+                len += sprintf(buf+len, "%-89s\n",
+                               "\nVirtual Service\n"
+                               "Pro VirtService        Conns     InPkts    OutPkts    InBytes   OutBytes");
+        }
+        
+        for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
+                l = &ip_vs_svc_table[idx];
+                for (e=l->next; e!=l; e=e->next) {
+                        svc = list_entry(e, struct ip_vs_service, s_list);
+                        pos += 80;
+                        if (pos <= offset)
+                                continue;
+                        if (pos > offset+length) {
+                                pos -= 80;
+                                goto done;
+                        }
+                        spin_lock(&svc->stats.lock);
+                        sprintf(temp, "%3s %08X:%04X %10lu %10lu %10lu %10lu %10lu",
+                                ip_vs_proto_name(svc->protocol),
+                                ntohl(svc->addr),
+                                ntohs(svc->port),
+                                svc->stats.conns,
+                                svc->stats.inpkts,
+                                svc->stats.outpkts,
+                                svc->stats.inbytes,
+                                svc->stats.outbytes);
+                        spin_unlock(&svc->stats.lock);
+                        len += sprintf(buf+len, "%-79s\n", temp);
+                }
+        }
+
+        for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
+                l = &ip_vs_svc_fwm_table[idx];
+                for (e=l->next; e!=l; e=e->next) {
+                        svc = list_entry(e, struct ip_vs_service, f_list);
+                        pos += 80;
+                        if (pos <= offset)
+                                continue;
+                        if (pos > offset+length) {
+                                pos -= 80;
+                                goto done;
+                        }
+                        spin_lock(&svc->stats.lock);
+                        sprintf(temp, "FWM %-13X %10lu %10lu %10lu %10lu %10lu",
+                                svc->fwmark,
+                                svc->stats.conns,
+                                svc->stats.inpkts,
+                                svc->stats.outpkts,
+                                svc->stats.inbytes,
+                                svc->stats.outbytes);
+                        spin_unlock(&svc->stats.lock);
+                        len += sprintf(buf+len, "%-79s\n", temp);
+                }
+        }
+
+        /* print the real server statistics */
+        pos += 105;
+        if (pos > offset+length) {
+                pos -= 105;
+                goto done;
+        }
+        if (pos > offset) {
+                len += sprintf(buf+len, "%-104s\n",
+                               "\nReal Service\n"
+                               "Pro VirtService   RealService        Conns     InPkts    OutPkts    InBytes   OutBytes");
+        }
+        
+        for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
+                l = &ip_vs_svc_table[idx];
+                for (e=l->next; e!=l; e=e->next) {
+                        svc = list_entry(e, struct ip_vs_service, s_list);
+                        p = &svc->destinations;
+                        for (q=p->next; q!=p; q=q->next) {
+                                dest = list_entry(q, struct ip_vs_dest, n_list);
+                                pos += 90;
+                                if (pos <= offset)
+                                        continue;
+                                if (pos > offset+length) {
+                                        pos -= 90;
+                                        goto done;
+                                }
+                                spin_lock(&dest->stats.lock);
+                                sprintf(temp,
+                                        "%3s %08X:%04X %08X:%04X %10lu %10lu %10lu %10lu %10lu",
+                                        ip_vs_proto_name(svc->protocol),
+                                        ntohl(svc->addr),
+                                        ntohs(svc->port),
+                                        ntohl(dest->addr),
+                                        ntohs(dest->port),
+                                        dest->stats.conns,
+                                        dest->stats.inpkts,
+                                        dest->stats.outpkts,
+                                        dest->stats.inbytes,
+                                        dest->stats.outbytes);
+                                spin_unlock(&dest->stats.lock);
+                                len += sprintf(buf+len, "%-89s\n", temp);
+                        }
+                }
+        }
+
+        for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
+                l = &ip_vs_svc_fwm_table[idx];
+                for (e=l->next; e!=l; e=e->next) {
+                        svc = list_entry(e, struct ip_vs_service, f_list);
+                        p = &svc->destinations;
+                        for (q=p->next; q!=p; q=q->next) {
+                                dest = list_entry(q,struct ip_vs_dest,n_list);
+                                pos += 90;
+                                if (pos <= offset)
+                                        continue;
+                                if (pos > offset+length) {
+                                        pos -= 90;
+                                        goto done;
+                                }
+                                spin_lock(&dest->stats.lock);
+                                sprintf(temp,
+                                        "FWM %-13X %08X:%04X %10lu %10lu %10lu %10lu %10lu",
+                                        svc->fwmark,
+                                        ntohl(dest->addr),
+                                        ntohs(dest->port),
+                                        dest->stats.conns,
+                                        dest->stats.inpkts,
+                                        dest->stats.outpkts,
+                                        dest->stats.inbytes,
+                                        dest->stats.outbytes);
+                                spin_unlock(&dest->stats.lock);
+                                len += sprintf(buf+len, "%-89s\n", temp);
+                        }
+                }
+        }
+  done:
+        read_unlock_bh(&__ip_vs_svc_lock);
+        
+        *start = buf+len-(pos-offset);          /* Start of wanted data */
+        len = pos-offset;
+        if (len > length)
+                len = length;
+        if (len < 0)
+                len = 0;
         return len;
 }
 
@@ -1542,15 +1736,12 @@ do_ip_vs_set_ctl(struct sock *sk, int cmd, void *user, unsigned int len)
 	else
 		svc = __ip_vs_svc_fwm_get(urule->vfwmark);
 
-	if (svc)
-                ip_vs_service_put(svc);
-
 	switch (cmd) {
 	case IP_VS_SO_SET_ADD:
 		if (svc != NULL)
 			ret = -EEXIST;
 		else
-			ret = ip_vs_add_service(urule);
+			ret = ip_vs_add_service(urule, &svc);
 		break;
 	case IP_VS_SO_SET_EDIT:
 		if (svc == NULL || svc->protocol != urule->protocol)
@@ -1561,8 +1752,11 @@ do_ip_vs_set_ctl(struct sock *sk, int cmd, void *user, unsigned int len)
 	case IP_VS_SO_SET_DEL:
 		if (svc == NULL || svc->protocol != urule->protocol)
 			ret =  -ESRCH;
-		else
+		else {
 			ret = ip_vs_del_service(svc);
+                        if (!ret)
+                                svc = NULL;
+                }
 		break;
 	case IP_VS_SO_SET_ADDDEST:
 		if (svc == NULL || svc->protocol != urule->protocol)
@@ -1587,6 +1781,9 @@ do_ip_vs_set_ctl(struct sock *sk, int cmd, void *user, unsigned int len)
 		break;
 	}
 
+	if (svc)
+                ip_vs_service_put(svc);
+
   out_unlock:
 	spin_unlock(&__ip_vs_mutex);
 	MOD_DEC_USE_COUNT;
@@ -1638,6 +1835,7 @@ int __init ip_vs_control_init(void)
         }
 
         proc_net_create("ip_vs", 0, ip_vs_get_info);
+        proc_net_create("ip_vs_stats", 0, ip_vs_stats_get_info);
 
         ipv4_vs_table.sysctl_header =
                 register_sysctl_table(ipv4_vs_table.root_dir, 0);
@@ -1661,6 +1859,7 @@ void __exit ip_vs_control_cleanup(void)
 {
         EnterFunction(2);
         unregister_sysctl_table(ipv4_vs_table.sysctl_header);
+        proc_net_remove("ip_vs_stats");
         proc_net_remove("ip_vs");
         nf_unregister_sockopt(&ip_vs_sockopts);
         LeaveFunction(2);
